@@ -10,16 +10,18 @@ import os
 from pathlib import Path
 from typing import Any, Optional
 
+import h5py
 import joblib
 from omegaconf import DictConfig
 import pandas as pd
+from rich.console import Console
 from rich.table import Table
 import wandb
 import xarray as xr
 
 from l2hmc.configs import AnnealingSchedule, Steps
-from l2hmc.utils.console import console
 from l2hmc.utils.plot_helpers import make_ridgeplots, plot_dataArray
+from l2hmc.utils.rich import is_interactive
 
 os.environ['AUTOGRAPH_VERBOSITY'] = '0'
 log = logging.getLogger(__name__)
@@ -59,19 +61,52 @@ def setup_annealing_schedule(cfg: DictConfig) -> AnnealingSchedule:
 def save_dataset(
         dataset: xr.Dataset,
         outdir: os.PathLike,
-        job_type: str = None,
+        use_hdf5: Optional[bool] = True,
+        job_type: Optional[str] = None,
+        **kwargs,
 ) -> Path:
-    fname = 'dataset.nc' if job_type is None else f'{job_type}_dataset.nc'
-    datafile = Path(outdir).joinpath(fname)
-    mode = 'a' if datafile.is_file() else 'w'
-    log.info(f'Saving dataset to: {datafile.as_posix()}')
-    datafile.parent.mkdir(exist_ok=True, parents=True)
-    dataset.to_netcdf(datafile.as_posix(), mode=mode)
+    if use_hdf5:
+        fname = 'dataset.h5' if job_type is None else f'{job_type}_data.h5'
+        outfile = Path(outdir).joinpath(fname)
+        dataset_to_h5pyfile(outfile, dataset=dataset, **kwargs)
+    else:
+        fname = 'dataset.nc' if job_type is None else f'{job_type}_dataset.nc'
+        outfile = Path(outdir).joinpath(fname)
+        mode = 'a' if outfile.is_file() else 'w'
+        log.info(f'Saving dataset to: {outfile.as_posix()}')
+        outfile.parent.mkdir(exist_ok=True, parents=True)
+        dataset.to_netcdf(outfile.as_posix(), mode=mode)
 
-    return datafile
+    return outfile
 
 
-def table_to_dict(table: Table, data: dict = None) -> dict:
+def dataset_to_h5pyfile(hfile: os.PathLike, dataset: xr.Dataset, **kwargs):
+    f = h5py.File(hfile, 'a')
+    for key, val in dataset.data_vars.items():
+        arr = val.values
+        if len(arr) == 0:
+            continue
+        if key in list(f.keys()):
+            shape = (f[key].shape[0] + arr.shape[0])  # type: ignore
+            f[key].resize(shape, axis=0)              # type: ignore
+            f[key][-arr.shape[0]:] = arr              # type: ignore
+        else:
+            maxshape = (None,)
+            if len(arr.shape) > 1:
+                maxshape = (None, *arr.shape[1:])
+            f.create_dataset(key, data=arr, maxshape=maxshape, **kwargs)
+
+    f.close()
+
+
+def dataset_from_h5pyfile(hfile: os.PathLike) -> dict:
+    f = h5py.File(hfile, 'r')
+    data = {key: f[key] for key in list(f.keys())}
+    f.close()
+    return data
+
+
+def table_to_dict(table: Table, data: Optional[dict] = None) -> dict:
     if data is None:
         return {
             column.header: [
@@ -95,9 +130,9 @@ def table_to_dict(table: Table, data: dict = None) -> dict:
 def save_logs(
         tables: dict[str, Table],
         summaries: Optional[list[str]] = None,
-        job_type: str = None,
+        job_type: Optional[str] = None,
         # rows: Optional[dict] = None,  # type:ignore
-        logdir: os.PathLike = None,
+        logdir: Optional[os.PathLike] = None,
         run: Optional[Any] = None,
 ) -> None:
     job_type = 'job' if job_type is None else job_type
@@ -106,10 +141,10 @@ def save_logs(
     else:
         logdir = Path(logdir)
 
-    cfile = logdir.joinpath('console.txt').as_posix()
-    text = console.export_text()
-    with open(cfile, 'w') as f:
-        f.write(text)
+    # cfile = logdir.joinpath('console.txt').as_posix()
+    # text = console.export_text()
+    # with open(cfile, 'w') as f:
+    #     f.write(text)
 
     table_dir = logdir.joinpath('tables')
     tdir = table_dir.joinpath('txt')
@@ -121,8 +156,8 @@ def save_logs(
     tfile = tdir.joinpath('table.txt')
     tfile.parent.mkdir(exist_ok=True, parents=True)
 
-    # data = {}
     data = {}
+    console = Console(record=True)
     for idx, table in tables.items():
         if idx == 0:
             data = table_to_dict(table)
@@ -168,10 +203,10 @@ def make_subdirs(basedir: os.PathLike):
 
 def plot_dataset(
         dataset: xr.Dataset,
-        nchains: int = 10,
-        outdir: os.PathLike = None,
-        title: str = None,
-        job_type: str = None,
+        nchains: Optional[int] = 10,
+        outdir: Optional[os.PathLike] = None,
+        title: Optional[str] = None,
+        job_type: Optional[str] = None,
         # run: Any = None,
 ) -> None:
     outdir = Path(outdir) if outdir is not None else Path(os.getcwd())
@@ -209,17 +244,19 @@ def plot_dataset(
     _ = make_ridgeplots(dataset,
                         outdir=outdir,
                         drop_nans=True,
+                        drop_zeros=False,
                         num_chains=nchains)
 
 
 def analyze_dataset(
         dataset: xr.Dataset,
         outdir: os.PathLike,
-        nchains: int = 16,
-        title: str = None,
-        job_type: str = None,
-        save: bool = True,
-        run: Any = None,
+        nchains: Optional[int] = 16,
+        title: Optional[str] = None,
+        job_type: Optional[str] = None,
+        save: Optional[bool] = True,
+        run: Optional[Any] = None,
+        use_hdf5: Optional[bool] = True,
 ):
     job_type = job_type if job_type is not None else f'job-{get_timestamp()}'
     dirs = make_subdirs(outdir)
@@ -231,6 +268,7 @@ def analyze_dataset(
     if save:
         try:
             datafile = save_dataset(dataset,
+                                    use_hdf5=use_hdf5,
                                     outdir=dirs['data'],
                                     job_type=job_type)
         except ValueError:
@@ -255,3 +293,39 @@ def analyze_dataset(
             run.log_artifact(artifact)
 
     return dataset
+
+
+def save_and_analyze_data(
+        dataset: xr.Dataset,
+        outdir: os.PathLike,
+        run: Optional[Any] = None,
+        output: Optional[dict] = None,
+        nchains: Optional[int] = -1,
+        job_type: Optional[str] = None,
+        framework: Optional[str] = None,
+) -> xr.Dataset:
+    jstr = f'{job_type}'
+    output = {} if output is None else output
+    title = (
+        jstr if framework is None
+        else ': '.join([jstr, f'{framework}'])
+    )
+
+    dataset = analyze_dataset(dataset,
+                              run=run,
+                              save=True,
+                              outdir=outdir,
+                              nchains=nchains,
+                              job_type=job_type,
+                              title=title)
+    if not is_interactive():
+        edir = Path(outdir).joinpath('logs')
+        edir.mkdir(exist_ok=True, parents=True)
+        log.info(f'Saving {job_type} logs to: {edir.as_posix()}')
+        save_logs(run=run,
+                  logdir=edir,
+                  job_type=job_type,
+                  tables=output.get('tables', None),
+                  summaries=output.get('summaries'))
+
+    return dataset
